==> TundereBird-0.23.tar.bz2: TundereBird-0.23/core/Smdn.Net.Imap4.Client/Smdn.Net.Imap4.WebClients/ImapFetchMessageBodyStream.cs <==
// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2008-2010 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.IO;
using System.Net;

using Smdn.Net.Imap4.Client.Session;

namespace Smdn.Net.Imap4.WebClients {
  internal class ImapFetchMessageBodyStream : Stream {
    public override bool CanSeek {
      get { RejectDisposed(); return false; }
    }

    public override bool CanRead {
      get { RejectDisposed(); return true; }
    }

    public override bool CanWrite {
      get { RejectDisposed(); return false; }
    }

    public override bool CanTimeout {
      get { RejectDisposed(); return true; }
    }

    public override long Position {
      get { RejectDisposed(); return position; }
      set { throw new NotSupportedException(); }
    }

    public override long Length {
      get { RejectDisposed(); return length; }
    }

    public override int ReadTimeout {
      get { RejectDisposed(); return session.ReceiveTimeout; }
      set { RejectDisposed(); session.ReceiveTimeout = value; }
    }

    public override int WriteTimeout {
      get { RejectDisposed(); return session.SendTimeout; }
      set { RejectDisposed(); session.SendTimeout = value; }
    }

    internal ImapFetchMessageBodyStream(ImapSession session, bool keepAlive, long fetchUid, string fetchSection, long rfc822size, int fetchPartialSize, int readWriteTimeout)
    {
      this.buffer = new byte[fetchPartialSize];
      this.session = session;
      this.keepAlive = keepAlive;
      this.fetchUid = fetchUid;
      this.fetchSection = fetchSection;
      this.length = rfc822size;

      ReadTimeout   = readWriteTimeout;
      WriteTimeout  = readWriteTimeout;
    }

    public override void Close()
    {
      Dispose(true);
    }

    protected override void Dispose(bool disposing)
    {
      base.Dispose(disposing);

      buffer = null;

      if (session != null) {
        if (!keepAlive)
          ImapWebRequest.CloseSession(session);

        session = null;
      }
    }

    public override void SetLength(long value)
    {
      throw new NotSupportedException();
    }

    public override long Seek(long offset, SeekOrigin origin)
    {
      throw new NotSupportedException();
    }

    public override void Flush()
    {
      throw new NotSupportedException();
    }

    public override void Write(byte[] buffer, int offset, int count)
    {
      throw new NotSupportedException();
    }

    public override int Read(byte[] dest, int offset, int count)
    {
      RejectDisposed();

      if (dest == null)
        throw new ArgumentNullException("dest");
      if (offset < 0)
        throw new ArgumentOutOfRangeException("offset", "must be greater than or equals to 0");
      if (count < 0)
        throw new ArgumentOutOfRangeException("count", "must be greater than or equals to 0");
      if (dest.Length < offset + count)
        throw new ArgumentException("invalid range");

      var read = 0;

      for (;;) {
        if (bufRemain == 0) {
          if (eof)
            return read;
          else
            eof = FillBuffer();
        }

        if (count <= bufRemain) {
          Buffer.BlockCopy(buffer, bufOffset, dest, offset, count);

          bufOffset += count;
          bufRemain -= count;
          position  += count;

          read += count;

          return read;
        }
        else {
          Buffer.BlockCopy(buffer, bufOffset, dest, offset, bufRemain);

          offset    += bufRemain;
          count     -= bufRemain;
          position  += bufRemain;

          read += bufRemain;

          bufRemain = 0;
        }
      }
    }

    private bool FillBuffer()
    {
      ImapMessage[] fetchPartial;

      try {
        var result = session.Fetch(ImapSequenceSet.CreateUidSet(fetchUid),
                                   ImapFetchDataItem.BodyPeek(fetchSection, position, buffer.Length),
                                   out fetchPartial);

        if (result.Failed)
          throw new WebException(result.Description, WebExceptionStatus.ReceiveFailure);
      }
      catch (TimeoutException ex) {
        throw new WebException("timed out", ex, WebExceptionStatus.Timeout, null);
      }
      catch (ImapException ex) {
        throw new WebException("unexpected error", ex, WebExceptionStatus.UnknownError, null);
      }

      Buffer.BlockCopy(fetchPartial[0].Body, 0, buffer, 0, fetchPartial[0].Body.Length);

      bufOffset = 0;
      bufRemain = fetchPartial[0].Body.Length;

      return (fetchPartial[0].Body.Length < buffer.Length);
    }

    private void RejectDisposed()
    {
      if (session == null)
        throw new ObjectDisposedException(GetType().Name);
    }

    private /*readonly*/ ImapSession session;
    private readonly bool keepAlive;
    private readonly long fetchUid;
    private /*readonly*/ string fetchSection;
    private readonly long length;
    private /*readonly*/ byte[] buffer;
    private long position = 0;
    private int bufOffset = 0;
    private int bufRemain  = 0;
    private bool eof = false;
  }
}

==> TundereBird-0.24.tar.bz2: TundereBird-0.24/core/Smdn.Net.Imap4.Client/Smdn.Net.Imap4.WebClients/ImapFetchMessageBodyStream.cs <==
// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2008-2010 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.IO;
using System.Net;

using Smdn.Net.Imap4.Client.Session;
using Smdn.Net.Imap4.Client.Transaction;

namespace Smdn.Net.Imap4.WebClients {
  internal class ImapFetchMessageBodyStream : Stream {
    public override bool CanSeek {
      get { RejectDisposed(); return false; }
    }

    public override bool CanRead {
      get { RejectDisposed(); return true; }
    }

    public override bool CanWrite {
      get { RejectDisposed(); return false; }
    }

    public override bool CanTimeout {
      get { RejectDisposed(); return true; }
    }

    public override long Position {
      get { RejectDisposed(); return position; }
      set { throw new NotSupportedException(); }
    }

    public override long Length {
      get { RejectDisposed(); return length; }
    }

    public override int ReadTimeout {
      get { RejectDisposed(); return session.ReceiveTimeout; }
      set { RejectDisposed(); session.ReceiveTimeout = value; }
    }

    public override int WriteTimeout {
      get { RejectDisposed(); return session.SendTimeout; }
      set { RejectDisposed(); session.SendTimeout = value; }
    }

    internal ImapFetchMessageBodyStream(ImapSession session, bool keepAlive, ImapSequenceSet fetchUidSet, int fetchBlockSize, int readWriteTimeout)
      : this(session, keepAlive, fetchUidSet, null, null, fetchBlockSize, readWriteTimeout)
    {
      partialFetch = false;
    }

    internal ImapFetchMessageBodyStream(ImapSession session, bool keepAlive, ImapSequenceSet fetchUidSet, string fetchSection, ImapPartialRange fetchRange, int fetchBlockSize, int readWriteTimeout)
    {
      partialFetch = true;

      this.buffer = new byte[fetchBlockSize];
      this.session = session;
      this.keepAlive = keepAlive;
      this.fetchUidSet = fetchUidSet;
      this.fetchSection = fetchSection;
      this.fetchRange = fetchRange;

      ReadTimeout   = readWriteTimeout;
      WriteTimeout  = readWriteTimeout;
    }

    public override void Close()
    {
      Dispose(true);
    }

    protected override void Dispose(bool disposing)
    {
      base.Dispose(disposing);

      buffer = null;

      if (session != null) {
        if (!keepAlive)
          ImapWebRequest.CloseSession(session);

        session = null;
      }
    }

    public override void SetLength(long value)
    {
      throw new NotSupportedException();
    }

    public override long Seek(long offset, SeekOrigin origin)
    {
      throw new NotSupportedException();
    }

    public override void Flush()
    {
      throw new NotSupportedException();
    }

    public override void Write(byte[] buffer, int offset, int count)
    {
      throw new NotSupportedException();
    }

    public override int Read(byte[] dest, int offset, int count)
    {
      RejectDisposed();

      if (dest == null)
        throw new ArgumentNullException("dest");
      if (offset < 0)
        throw new ArgumentOutOfRangeException("offset", "must be greater than or equals to 0");
      if (count < 0)
        throw new ArgumentOutOfRangeException("count", "must be greater than or equals to 0");
      if (dest.Length < offset + count)
        throw new ArgumentException("invalid range");

      var read = 0;

      for (;;) {
        if (bufRemain == 0) {
          if (eof)
            return read;
          else
            eof = FillBuffer();
        }

        if (count <= bufRemain) {
          Buffer.BlockCopy(buffer, bufOffset, dest, offset, count);

          bufOffset += count;
          bufRemain -= count;
          position  += count;

          read += count;

          return read;
        }
        else {
          Buffer.BlockCopy(buffer, bufOffset, dest, offset, bufRemain);

          offset    += bufRemain;
          count     -= bufRemain;
          position  += bufRemain;

          read += bufRemain;

          bufRemain = 0;
        }
      }
    }

    internal ImapTransactionResult Prepare(out ImapMessage messageAttr)
    {
      messageAttr = null;

      ImapMessage[] fetched;
      ImapFetchDataItem fetchDataItem;

      if (partialFetch) {
        var fetchStart = (fetchRange == null) ? 0L : fetchRange.Start;
        var fetchLength = (fetchRange == null || !fetchRange.IsLengthSpecified)
          ? buffer.Length
          : Math.Min(fetchRange.Length.Value, (long)buffer.Length);

        fetchDataItem = ImapFetchDataItem.BodyPeek(fetchSection, fetchStart, fetchLength);
      }
      else {
        fetchDataItem = ImapFetchDataItem.Full + ImapFetchDataItem.BodyPeek(0L, buffer.Length);
      }

      var result = session.Fetch(fetchUidSet,
                                 fetchDataItem,
                                 out fetched);

      if (result.Succeeded) {
        if (fetched.Length == 0)
          throw new WebException("no such message", null, WebExceptionStatus.Success, new ImapWebResponse(result));

        messageAttr = fetched[0];

        if (messageAttr.Body == null)
          throw new WebException("empty body", null, WebExceptionStatus.Success, new ImapWebResponse(result));

        Buffer.BlockCopy(messageAttr.Body, 0, buffer, 0, messageAttr.Body.Length);

        bufOffset = 0;
        bufRemain = messageAttr.Body.Length;

        eof = (messageAttr.Body.Length < buffer.Length);

        if (!partialFetch)
          length = messageAttr.Rfc822Size;
      }
      else {
        // not to close session by this instance
        session = null;
      }

      return result;
    }

    private bool FillBuffer()
    {
      ImapMessage[] fetched;

      try {
        ImapFetchDataItem fetchDataItem;

        if (partialFetch) {
        var fetchStart = (fetchRange == null) ? 0L : fetchRange.Start;
        var fetchLength = (fetchRange == null || !fetchRange.IsLengthSpecified)
          ? buffer.Length
          : Math.Min(fetchRange.Length.Value - position, (long)buffer.Length);

          fetchDataItem = ImapFetchDataItem.BodyPeek(fetchSection, position + fetchStart, fetchLength);
        }
        else {
          fetchDataItem = ImapFetchDataItem.BodyPeek(position, buffer.Length);
        }

        var result = session.Fetch(fetchUidSet,
                                   fetchDataItem,
                                   out fetched);

        if (result.Failed)
          throw new WebException(result.Description, WebExceptionStatus.ReceiveFailure);
      }
      catch (TimeoutException ex) {
        throw new WebException("timed out", ex, WebExceptionStatus.Timeout, null);
      }
      catch (ImapException ex) {
        throw new WebException("unexpected error", ex, WebExceptionStatus.UnknownError, null);
      }

      Buffer.BlockCopy(fetched[0].Body, 0, buffer, 0, fetched[0].Body.Length);

      bufOffset = 0;
      bufRemain = fetched[0].Body.Length;

      return (fetched[0].Body.Length < buffer.Length);
    }

    private void RejectDisposed()
    {
      if (session == null)
        throw new ObjectDisposedException(GetType().Name);
    }

    private /*readonly*/ ImapSession session;
    private readonly bool keepAlive;
    private readonly bool partialFetch;
    private ImapSequenceSet fetchUidSet;
    private string fetchSection;
    private ImapPartialRange fetchRange;
    private long length;
    private /*readonly*/ byte[] buffer;
    private long position = 0;
    private int bufOffset = 0;
    private int bufRemain  = 0;
    private bool eof = false;
  }
}
